import os  # os.split
import sys  # sys.stdout
import telnetlib  # telnet connection to gns3 routers
import time  # time.sleep
import argparse
import json  # read .gns3
from multiprocessing.dummy import Pool as ThreadPool  # threads used to get confs
import threading

# GLOBAL VARS #############################

# Empty list used to return gns3 routers output
confs_routeurs = []

#Used to clean configurations
mpls_filter = ["mpls ip", "%LDP-5-NBRCHG"]
bgp_filter = ["no synchronization", "bgp log-neighbor-changes", "no auto-summary", "%LDP-5-NBRCHG"]

router_filter = ["CE1", "CE2", "CE3", "RRF"]

# router name <=> port correspondance
dict_name_port = {}

# Server running gns3
routeurs_addr = "127.0.0.1"

# GNS3 Project paths
confs_file_path = "Cisco_topology2"
gns_file = "Cisco_topology2/Cisco_topology2.gns3"

# Options
verbose = False
latex = False
threadlock = threading.Lock()

# END GLOBAL VAR ###########################


# TELNET FUNCTIONS

# Clean telnet console
def flush(tn):
    tn.write("\r\nend\r\nq\r\n\r\nterminal length 0\r\n\r\n")
    time.sleep(5)
    tn.read_very_eager()


# delete useless informations from the retrieved configurations
def filter_info(info, filter):
    for x in filter:
        info = info.replace("{0}".format(x), "")
    info_lines = info.split("\n")
    new_contents = []
    for i in info_lines:
        if not i.strip() or i in filter:
            continue
        else:
            new_contents.append("{0}\n".format(i))

    return "".join(new_contents)

# get a configuration section
def get_section(name, tn, section):
    initiated_tn = False
    if tn is None:
        tn = telnetlib.Telnet(routeurs_addr, dict_name_port[name])
        initiated_tn = True
    flush(tn)
    tn.write("\r\n\r\nsh run | section {0}\r\n".format(section))
    time.sleep(15)
    out = tn.read_very_eager()
    processed_out = ""
    for line in out.split("\r"):
        if "#" not in line:
            processed_out = processed_out + line
    return processed_out  # processed_out

    if initiated_tn:
        tn.close()


# retrieve parts of the running configuration
def get_running_conf(name):
    if name in router_filter:
        return
    if verbose:
        sys.stdout.write("In thread\n")
    tn = telnetlib.Telnet(routeurs_addr, dict_name_port[name])
    flush(tn)

    # get IOS version
    version = get_section(name, tn, "version")
    # get mpls info
    mpls = get_section(name, tn, "mpls")
    mpls = filter_info(mpls, mpls_filter)

    # get bgp info
    bgp = get_section(name, tn, "bgp")
    bgp = filter_info(bgp, bgp_filter)

    conf_routeur = "%s%s\n\n%s\n%s\n" % (name, version, mpls, bgp)

    with threadlock:
        confs_routeurs.append(conf_routeur)
    tn.close()
    return conf_routeur



# set router ldp config to explicit null
def set_uhp(name):
    if verbose:
        sys.stdout.write("Setting UHP on {0}\n".format(name))
    tn = telnetlib.Telnet(routeurs_addr, dict_name_port[name])
    flush(tn)
    tn.write("\r\n conf t \r\n")
    tn.write("\r\n mpls ldp explicit-null \r\n")
    tn.write("\r\n end \r\n")
    time.sleep(15)
    tn.close()



# set router ldp config to implicit null
def set_php(name):
    if verbose:
        sys.stdout.write(
            "Setting PHP on {0}:{1}\n".format(name,
                                              dict_name_port[name]))
    tn = telnetlib.Telnet(routeurs_addr, dict_name_port[name])
    flush(tn)
    tn.write("\r\n conf t \r\n")
    tn.write("\r\n no mpls ldp explicit-null \r\n")
    tn.write("\r\n end \r\n")
    time.sleep(15)
    tn.close()


# set ip ttl propagation on a router
def set_prop(name):
    if verbose:
        sys.stdout.write(
            "Setting ip propagate on {0}:{1}\n".format(name,
                                                       dict_name_port[name]))
    tn = telnetlib.Telnet(routeurs_addr, dict_name_port[name])
    flush(tn)
    tn.write("\r\n conf t \r\n")
    tn.write("\r\n mpls ip propagate-ttl \r\n")
    tn.write("\r\n end \r\n")
    time.sleep(15)
    tn.close()

# set ip no propagate ttl on a router
def set_noprop(name):
    if verbose:
        sys.stdout.write("Setting no ip propagate on {0}\n".format(name))
    tn = telnetlib.Telnet(routeurs_addr, dict_name_port[name])
    flush(tn)
    tn.write("\r\n conf t \r\n")
    tn.write("\r\n no mpls ip propagate-ttl \r\n")
    tn.write("\r\n end \r\n")
    time.sleep(15)
    tn.close()

# conf cisco router to behave like Juniper router during LDP exchange
def set_dpr(name):
    tn = telnetlib.Telnet(routeurs_addr, dict_name_port[name])
    flush(tn)
    tn.write("\r\n conf t \r\n")
    tn.write("\r\n mpls ldp label \r\n")
    tn.write("\r\n allocate global host-routes \r\n")
    tn.write("\r\n end \r\n")
    time.sleep(15)
    tn.close()

# revert to classic cisco configuration
def set_brpr(name):
    tn = telnetlib.Telnet(routeurs_addr, dict_name_port[name])
    flush(tn)
    tn.write("\r\n conf t \r\n")
    tn.write("\r\n mpls ldp label \r\n")
    tn.write("\r\n no allocate global host-routes \r\n")
    tn.write("\r\n end \r\n")
    time.sleep(15)
    tn.close()

# END TELNET

# print router configuration
def print_confs(confs):
    confs = sorted(confs)
    second_half = False
    if latex:
        sys.stdout.write("\\begin{minipage}{0.5\\textwidth}" + '\n')
        sys.stdout.write("\\begin{verbatim}")
        for i in xrange(0, len(confs)):
            if i < len(confs) / 2 +1 :
                sys.stdout.write(confs[i])
                sys.stdout.write("--\n\n")
            else:
                if not second_half:
                    sys.stdout.write("\\end{verbatim}")
                    sys.stdout.write("\\end{minipage}% \n \\hfill")
                    sys.stdout.write("\\begin{minipage}{0.5\\textwidth}")
                    sys.stdout.write("\\begin{tabular}{|p{\\textwidth}}\n")
                    sys.stdout.write("\\begin{verbatim}")
                    second_half = True
                sys.stdout.write(confs[i])
                sys.stdout.write("----\n\n")
        sys.stdout.write("\\end{verbatim}")
        sys.stdout.write("\\end{tabular}")
        sys.stdout.write("\\end{minipage}%\n")
        return

    for conf in confs:
        sys.stdout.write("\n=========================\n")
        sys.stdout.write(conf)
        sys.stdout.write("\n=========================\n")





# create port/name correspondance
def name_port():
    if verbose:
        sys.stdout.write("Console Port  <-> Name ( building dictionnary )\n")
    with open(gns_file) as gns_fd:
        gns_json = json.load(gns_fd)
        for nodes in gns_json['topology']['nodes']:
            if nodes['console'] is not None and nodes['console'] != "null":
                dict_name_port[nodes['name']] = int(nodes['console'])
                if verbose:
                    sys.stdout.write(
                        "{0} {1}\n".format(nodes['console'], nodes['name']))

# END MISC


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument("-path", type=str)
    parser.add_argument("-verbose", action='store_true')
    
    parser.add_argument("-uhp", nargs='*')
    parser.add_argument("-php", nargs='*')
    parser.add_argument("-no_prop", nargs='*')
    parser.add_argument("-prop", nargs='*')
    parser.add_argument("-dpr", nargs="*")
    parser.add_argument("-brpr", nargs="*")

    args = parser.parse_args()

    if args.path is not None:
        gns_file = args.path
        head, _ = os.path.split(gns_file)
        confs_file_path = head

    if args.verbose is True:
        verbose = True
        
    # port <-> name 
    name_port()

	

    # Launch threads
    number_of_threads = len(dict_name_port.keys())
    pool = ThreadPool(number_of_threads)
    

    if args.dpr is not None:
        pool_args = []
        for name in args.dpr:
            pool_args.append(name)       
        res = pool.map(set_dpr, pool_args)
        time.sleep(5)

    if args.brpr is not None:
        pool_args = []
        for name in args.brpr:
            pool_args.append(name)
        res = pool.map(set_brpr, pool_args)
        time.sleep(5)


    # set routers to UHP
    if args.uhp is not None:
        pool_args = []
        for router in args.uhp:
            pool_args.append(router)
        res = pool.map(set_uhp, pool_args)
        time.sleep(5)

	#set PHP
    if args.php is not None:
        pool_args = []
        for router in args.php:
            pool_args.append(router)
        res = pool.map(set_php, pool_args)
        time.sleep(5)

	#set no propagate
    if args.no_prop is not None:
        pool_args = []
        for router in args.no_prop:
            pool_args.append(router)
        res = pool.map(set_noprop, pool_args)
        time.sleep(5)

	#set propagate
    if args.prop is not None:
        pool_args = []
        for router in args.prop:
            pool_args.append(router)
        res = pool.map(set_prop, pool_args)
        time.sleep(5)
    

